
//Copyright (C) 1998-2005 MKTMK Moscow, Russia
//e-mail: mktmk<at>narod.ru
//interpreter Open Basic (OB).
//v.1.50 23.08.2003(last modification 18.12.2003)

//you may use Open Basic for:
// - script language expandable by user without compilation need
// - configuration file language
// - input language for terminal program (batch file is obtained automatically)
// - term paper for third year student of computer science

//features:
// - Open Basic (OB) is fully interpreter
// - OB is a C++ class with member-function (load, run etc.)
// - OB designed for embed to user application
// - OB distributed as a library or source
// - OB have Basic input language, expandable by user
// - OB support float, integer and string variable and array of this type
// - user may attach (connect) user function to Open Basic execution system
// - user function may be C/C++, ASM and other language
// - user function may be attached during run-time before run Basic program
// - user function may be call from Basic program as Basic function
// - user function may receive data from Basic program
// - user function may return data to Basic program
// - user function have access to all variable and array of Basic by name
// - user function may read and write variable and array of Basic program
// - OB written entirely in ANSI C++ without any compiler extention
// - OB have multithread-safe code
// - OB use standart C++ header file and streams
// - OB do not use any graphic libraries
// - OB do not use any virtual machine
// - OB use exception for error handling (for eliminate memory leak)
// - OB use template and virtual function
// - OB is OS independent
// - source code size of OB less then 150K
// - x86.exe code size of OB less then 100K if other libraries is dynamic link

#ifndef OB_THIS
#define OB_THIS

//common includes
#ifdef _MSC_VER

//MSVC7

#include <iostream>
#include <strstream>
#include <fstream>
#include <math.h>
#include <stdlib.h>
#include <ctype.h>
#include <limits.h>
#include <time.h>

#else

//BCB60 and GCC 3.2.2

#include <iostream.h>
#include <strstream.h>
#include <fstream.h>
#include <math.h>
#include <stdlib.h>
#include <ctype.h>
#include <limits.h>
#include <time.h>

#endif

using namespace std;

typedef ostrstream ob_ostrstream;
typedef strstream  ob_strstream;
typedef istrstream ob_istrstream;

//compilers condition
#ifdef __BORLANDC__
#define MSTORE_INLINE
#define MVECT_INLINE
#define MLIST_INLINE
#define MSTACK_INLINE
#define MHASH_INLINE
#define OB_INLINE
#endif

//containers for ob
#include <mstore.h>
#include <mvect.h>
#include <mlist.h>
#include <mstack.h>
#include <mhash.h>

//define
#ifndef OB_INLINE
#define OB_INLINE inline
#endif

#ifndef OB_NULL
#define OB_NULL (0)
#endif

//forward declarations
class ob_obasic;
class ob_par;
class ob_var;
class ob_int;
class ob_float;
class ob_str;

//types of Open Basic
//service types
typedef unsigned int      ob_type_arraydimension;       //1
typedef unsigned int      ob_type_arraysize;            //2
typedef streampos         ob_type_streampos;            //3
typedef unsigned int      ob_type_parnum;               //4
typedef unsigned int      ob_type_filenum;              //5
typedef unsigned char     ob_type_hashtablelength;      //6
typedef unsigned int      ob_type_stringsize;           //7
typedef unsigned int      ob_type_codeerror;            //8
typedef unsigned long int ob_type_serialnum;            //9
typedef unsigned int      ob_type_version;              //10
typedef unsigned int      ob_type_countlist;            //11
typedef          char     ob_type_char;                 //12
typedef unsigned int      ob_type_statictablesize;      //13
typedef unsigned int      ob_type_countnestedoperator;  //14
typedef unsigned int      ob_type_srandargument;        //15  

typedef mstore_dynamicalloc<ob_type_char,ob_type_stringsize> ob_type_d;
typedef mvect_common<ob_type_d>                              ob_type_s;

typedef mstore_staticdatawithoutbuf<ob_type_char,ob_type_stringsize> ob_type_d1;
typedef mvect_common<ob_type_d1>                                     ob_type_s1;

typedef mstore_dynamicalloc<ob_type_arraysize,ob_type_arraydimension> ob_type_a;
typedef mvect_common<ob_type_a>                                     ob_type_dim;

//const of Open Basic
//max length string in input operator
const ob_type_stringsize      ob_maxlengthstring=4096;
const ob_type_parnum          ob_maxnumpar=64;
const ob_type_stringsize      ob_timedatasize=100;
const ob_type_arraydimension  ob_maxarraydimention=20;
const ob_type_statictablesize ob_maxnumoperator=25;

const ob_type_char ob_const_sharp     ='#';
const ob_type_char ob_const_at        ='@';
const ob_type_char ob_const_equ       ='=';
const ob_type_char ob_const_cr        ='\r';
const ob_type_char ob_const_lf        ='\n';
const ob_type_char ob_const_asterisk  ='*';
const ob_type_char ob_const_slash     ='/';
const ob_type_char ob_const_plus      ='+';
const ob_type_char ob_const_minus     ='-';
const ob_type_char ob_const_pow       ='^';
const ob_type_char ob_const_lp        ='(';
const ob_type_char ob_const_rp        =')';
const ob_type_char ob_const_sc        =';';
const ob_type_char ob_const_comma     =',';
const ob_type_char ob_const_lp1       ='<';
const ob_type_char ob_const_rp1       ='>';
const ob_type_char ob_const_quotes    ='\"';
const ob_type_char ob_const_quotes1   ='\'';
const ob_type_char ob_const_point     ='.';

const ob_type_char ob_const_0         ='0';
const ob_type_char ob_const_1         ='1';
const ob_type_char ob_const_2         ='2';
const ob_type_char ob_const_3         ='3';
const ob_type_char ob_const_4         ='4';
const ob_type_char ob_const_5         ='5';
const ob_type_char ob_const_6         ='6';
const ob_type_char ob_const_7         ='7';
const ob_type_char ob_const_8         ='8';
const ob_type_char ob_const_9         ='9';

const ob_type_char ob_const_a         ='a';
const ob_type_char ob_const_b         ='b';
const ob_type_char ob_const_c         ='c';
const ob_type_char ob_const_d         ='d';
const ob_type_char ob_const_e         ='e';
const ob_type_char ob_const_f         ='f';
const ob_type_char ob_const_A         ='A';
const ob_type_char ob_const_B         ='B';
const ob_type_char ob_const_C         ='C';
const ob_type_char ob_const_D         ='D';
const ob_type_char ob_const_E         ='E';
const ob_type_char ob_const_F         ='F';

const ob_type_char ob_const_space     =' ';
const ob_type_char ob_const_tab       ='\t';
const ob_type_char ob_const_ff        ='\f';

const ob_type_char ob_const_hex       ='x';
const ob_type_char ob_const_HEX       ='X';

const ob_type_char ob_const_underscore='_';
const ob_type_char ob_const_strlabel  ='$';
const ob_type_char ob_const_intlabel  ='%';

const ob_type_char ob_const_endstring =0;

//length of hash table
const ob_type_hashtablelength ob_hashlength=111;
const ob_type_hashtablelength ob_starthashvalue=1;

const ob_type_codeerror ob_basecodeerror=0;

//variable types
typedef float                    ob_type_flo;
typedef int                      ob_type_int;

class ob_timekeeper{
time_t* starttime;
time_t* stoptime;
public:
ob_timekeeper(time_t* startt,time_t* stopt):
starttime(startt),stoptime(stopt){time(starttime);}
~ob_timekeeper(){time(stoptime);}
};

class ob_type_string:public ob_type_s{
public:
ob_type_string(){}//for ob_str default constructor
ob_type_string(ob_type_stringsize c):ob_type_s(c){}//for current string in ob_lex
ob_type_string(ob_type_char* data1,ob_type_stringsize c):
ob_type_s(data1,c){}//for create string without new call strlen in prim()
ob_type_string(ob_type_char* data1)://for create string after return char* fron user func.
ob_type_s(data1,ob_type_char(),ob_maxlengthstring){}
};

//prototypes 
ob_type_hashtablelength ob_hash(ob_type_char* data,ob_type_stringsize l);

inline void ob_seekg(istream* in,ob_type_streampos pos){//clr eof flag
in->clear();
in->seekg(pos);
}

const ob_type_char*
ob_getstringparam(
const ob_type_parnum *descrc,
const ob_type_char* parc,
const ob_type_parnum i);

ob_type_stringsize ob_strlen(ob_type_char* str);

void ob_strcpy(ob_type_char* datadst,ob_type_char* datasrc);

void
ob_strncpy(ob_type_char* datadst,ob_type_char* datasrc,ob_type_stringsize max);

class ob_type_namestatic:public ob_type_s1{
public:
ob_type_namestatic():ob_type_s1(OB_NULL,ob_type_stringsize()){}//for member ob_lex
ob_type_namestatic(ob_type_char* data1):ob_type_s1(data1,ob_strlen(data1)){}//for findcommon
ob_type_namestatic(ob_type_char* data1,ob_type_stringsize c):ob_type_s1(data1,c){}//for get_lex

ob_type_hashtablelength gethash(){return ob_hash(data,count);}
};

class ob_type_name:public ob_type_s{
public:
ob_type_name(){}//for local ob_label
ob_type_name(const ob_type_name& name):ob_type_s(name.getdata(),name.getcount()){}//for ob_forfor
ob_type_name(const ob_type_namestatic& name):ob_type_s(name.getdata(),name.getcount()){}//for ob_label and ob_varbase
ob_type_name(ob_type_char* data1):ob_type_s(data1,ob_type_char(),ob_maxlengthstring){}
ob_type_hashtablelength gethash(){return ob_hash(data,count);}//for insert new var in tables
};

#define OB_CODEERROR(err)              \
const ob_type_codeerror ob_err##err=(ob_basecodeerror+err);

#define OB_CODEERRORGROUP(e0,e1,e2,e3,e4,e5,e6,e7,e8,e9) \
OB_CODEERROR(e0)                                         \
OB_CODEERROR(e1)                                         \
OB_CODEERROR(e2)                                         \
OB_CODEERROR(e3)                                         \
OB_CODEERROR(e4)                                         \
OB_CODEERROR(e5)                                         \
OB_CODEERROR(e6)                                         \
OB_CODEERROR(e7)                                         \
OB_CODEERROR(e8)                                         \
OB_CODEERROR(e9)

OB_CODEERRORGROUP( 0, 1, 2, 3, 4, 5, 6, 7, 8, 9)
OB_CODEERRORGROUP(10,11,12,13,14,15,16,17,18,19)
OB_CODEERRORGROUP(20,21,22,23,24,25,26,27,28,29)
OB_CODEERRORGROUP(30,31,32,33,34,35,36,37,38,39)
OB_CODEERRORGROUP(40,41,42,43,44,45,46,47,48,49)
OB_CODEERRORGROUP(50,51,52,53,54,55,56,57,58,59)
OB_CODEERRORGROUP(60,61,62,63,64,65,66,67,68,69)
OB_CODEERRORGROUP(70,71,72,73,74,75,76,77,78,79)
OB_CODEERRORGROUP(80,81,82,83,84,85,86,87,88,89)

void ob_errorgeneration(ob_type_codeerror e);//do not inline

#define OB_ERR(e) ob_errorgeneration(e);

//some local classes
//variable
enum ob_type_ident {OB_IDENTFLO,OB_IDENTINT,OB_IDENTSTR,OB_NOIDENT};

#define OB_DECLF(name,typearg) ob_var* name(typearg* a)

#define OB_DECLSETF(start,typearg,stop)                               \
start OB_DECLF(addv,typearg) stop;start OB_DECLF(subv,typearg) stop;  \
start OB_DECLF(mulv,typearg) stop;start OB_DECLF(divv,typearg) stop;  \
start OB_DECLF(pwwv,typearg) stop;start OB_DECLF(asgv,typearg) stop;  \
start bool                                                            \
compare(typearg* a,ob_var::ob_type_if currentlexif) stop;

#define OB_DECLSETALLF(nameclass,start,stop)                    \
OB_DECLSETF(start,ob_var,stop)                                  \
OB_DECLSETF(start,ob_float,stop)                                \
OB_DECLSETF(start,ob_int,stop)                                  \
OB_DECLSETF(start,ob_str,stop)                                  \
start ob_var* uniminus() stop;                                  \
start bool eq() const stop;                                     \
start ob_var* duplicate() stop;                                 \
start ob_type_ident isA() const stop;                           \
start void tostream(ostream* out) stop;                         \
start void topar(ob_par* p,const ob_type_parnum i) stop;        \
start ~nameclass(){}

//base class for types variable obasic
class ob_var{
public:
enum ob_type_if{LESSIF,GREATIF,EQVIF,LESSEQVIF,GREATEQVIF,NOTEQVIF};
OB_DECLSETALLF(ob_var,virtual,=0)
virtual void cast(ob_type_flo* a)=0;
virtual void cast(ob_type_int* a)=0;
virtual void cast(ob_type_char** a)=0;
virtual ob_type_stringsize size()=0;//for user information
};

//derived class for types variable obasic
class ob_float:public ob_var{
friend class ob_int;
friend class ob_str;

ob_type_flo var;
public:
typedef ob_type_flo typedatavar;

ob_float();
ob_float(ob_type_char*);
ob_float(ob_type_flo);

OB_DECLSETALLF(ob_float, , )

void cast(ob_type_flo* a);
void cast(ob_type_int* a);
void cast(ob_type_char** a);
ob_type_stringsize size(){return sizeof(var);}
};

class ob_int:public ob_var{
friend class ob_float;
friend class ob_str;

ob_type_int var;
public:
typedef ob_type_int typedatavar;

ob_int();
ob_int(ob_type_char*);
ob_int(ob_type_int);

OB_DECLSETALLF(ob_int, , )

void cast(ob_type_flo* a);
void cast(ob_type_int* a);
void cast(ob_type_char** a);
ob_type_stringsize size(){return sizeof(var);}
};

class ob_str:public ob_var{
friend class ob_float;
friend class ob_int;

ob_type_string var;
public:
typedef ob_type_char* typedatavar;

ob_str();//for array of string
ob_str(ob_type_char* s);//for return ob_type_char* from user function
ob_str(ob_type_char* s,ob_type_stringsize l);

OB_DECLSETALLF(ob_str, , )

void cast(ob_type_flo* a);
void cast(ob_type_int* a);
void cast(ob_type_char** a);
ob_type_stringsize size(){return var.getcount()-1;}
};

#define OB_OPERATIONWRAPPER(name)                         \
void name(ob_savewrappercommon* a){                       \
T* tmpptr=ptr->name##v(a->ptr);                           \
if(tmpptr==ptr){ a->clear();} else {clear();transfer(a);} \
}

//some realisation
template <class T>
class ob_savewrappercommon{
T* ptr;
ob_savewrappercommon(const ob_savewrappercommon& s):ptr(s.ptr){}
ob_savewrappercommon& operator =(const ob_savewrappercommon& s){}
public:
ob_savewrappercommon():ptr(OB_NULL){}
ob_savewrappercommon(T* p):ptr(p){}
void setptr(T* s){
if((ptr!=OB_NULL)&&(s!=OB_NULL)){
clear();//for eliminate double delete
OB_ERR(ob_err78)
}
ptr=s;
}
T* getptr() const {return ptr;}
void clear(){if(ptr!=OB_NULL) delete ptr;ptr=OB_NULL;}
~ob_savewrappercommon(){if(ptr!=OB_NULL) delete ptr;}

void transfer(ob_savewrappercommon* d){//for ob_forfor and vararray
setptr(d->ptr);
d->setptr(OB_NULL);
}

OB_OPERATIONWRAPPER(add)
OB_OPERATIONWRAPPER(sub)
OB_OPERATIONWRAPPER(mul)
OB_OPERATIONWRAPPER(div)
OB_OPERATIONWRAPPER(pww)

void asg(ob_savewrappercommon* a){
ptr->asgv(a->ptr);//do not check pointer because asgv always return own pointer
a->clear();
}
  
};

typedef ob_savewrappercommon<ob_var>       ob_savewrapper;

class ob_label:public ob_type_name{
protected:
ob_type_streampos label;
public:
ob_label(const ob_type_streampos p):label(p){}//local label

ob_label(ob_type_namestatic& name,const ob_type_streampos p)://label for table
ob_type_name(name),label(p){}

ob_label(ob_type_name& name,const ob_type_streampos p)://label for ob_forfor (i.e. bigname need)
ob_type_name(name),label(p){}

void posit(istream* in) const {ob_seekg(in,label);}

void set(ob_type_streampos p){label=p;}



virtual ~ob_label(){}
};

class ob_forfor:public ob_label{
ob_savewrapper lastvalue,step;
ob_var::ob_type_if typeifforcheckinfor;
public:

ob_forfor(ob_type_name& name,ob_type_streampos p,
ob_savewrapper* l,ob_savewrapper* s):ob_label(name,p){
lastvalue.transfer(l);
step.transfer(s);
ob_savewrapper zero(new ob_int(ob_type_int()));
if(step.getptr()->compare(zero.getptr(),ob_var::LESSIF)==false)
typeifforcheckinfor=ob_var::LESSEQVIF;
else
typeifforcheckinfor=ob_var::GREATEQVIF;
}

~ob_forfor(){}

void add(ob_savewrapper* var){
ob_savewrapper step1(step.getptr()->duplicate());
ob_savewrapper var1(var->getptr()->duplicate());
var1.add(&step1);
var->asg(&var1);
}

bool check(ob_savewrapper* var) const{
//return lastvalue.getptr()->compare(var->getptr(),typeifforcheckinfor);
return var->getptr()->compare(lastvalue.getptr(),typeifforcheckinfor);
}
};

//variable for store
class ob_varbase:public ob_type_name{
public:
ob_varbase(ob_type_namestatic& name):ob_type_name(name){}
virtual ob_type_name bigname()=0;//for FOR operator
virtual ob_savewrapper* get()=0;
virtual ~ob_varbase(){}//do not need virtual destructor because not delete varone and vararray as varbase
};

class ob_varone:public ob_varbase{
ob_savewrapper value;
public:

//ob_varone(ob_type_char* name);
ob_varone(ob_type_namestatic& name);

ob_type_ident isA(){return value.getptr()->isA();}

ob_savewrapper* get(){return (&value);}

ob_type_name bigname();//for FOR operator
~ob_varone(){}
};

typedef mstore_dynamicalloc<ob_savewrapper,ob_type_arraysize> ob_type_v;

//array for store
class ob_vararray:public ob_varbase{
mvect_common<ob_type_v> value;
ob_type_dim dimension;
ob_type_dim dimension1;
ob_type_arraysize sizecurrent;
ob_type_arraysize sizecreat;
public:

ob_vararray(ob_type_namestatic& name,ob_type_dim* ptrdim);

ob_type_ident isA(){
return value[0].getptr()->isA();
}

ob_savewrapper* get(){return &(value[sizecurrent]);}
void setsize(ob_type_dim* ptrdimcurrent);

const ob_type_arraysize* getsize(ob_type_arraydimension* kr){
*kr=dimension.getcount();
return dimension.c_str();
}

//return value because need modify name for FOR oper.
ob_type_name bigname();

virtual ~ob_vararray(){}
};

class ob_file{
fstream file;
ob_type_filenum number;
int modecreate;
public:

ob_file(
ob_type_char* name,
ios::openmode mode,
ob_type_filenum num):number(num){
file.open(name,mode);
}

~ob_file(){
file.close();
}
fstream* getfile(){return &file;}

ob_type_filenum getnum(){return number;}

};

class ob_lex;
typedef void (ob_lex::*ob_oper)();

class ob_operator:public ob_type_namestatic{
public:
enum flbreak{NOBREAK,BREAK};
private:
ob_oper ptroper;
flbreak before,after;
public:

ob_operator();//for trailer of operator table

ob_operator(ob_type_char* name,ob_oper p,flbreak b=NOBREAK,flbreak a=NOBREAK);

void run(ob_lex* lex);

void  setbreak(flbreak b=NOBREAK,flbreak a=NOBREAK){before=b;after=a;}
flbreak getbefore() const {return before;}
flbreak getafter()  const {return after;}
};

class ob_api:public ob_type_name{
public:
ob_api(ob_type_char* a):ob_type_name(a){}
virtual void run(ob_obasic* ptr,ob_par* p,ob_savewrapper* v)=0;
virtual ~ob_api(){}
};

template <class typecompare,class typestore> class ob_hashfind{
typecompare* d;
public:
ob_hashfind(typecompare& ddd):d(&ddd){}
ob_type_hashtablelength gethash(){return d->gethash();}
bool operator()(void* dd){
if(*((typestore*)dd)==(*d)) return true; else return false;
}
};

class ob_filefind{
const ob_type_filenum d;
public:
ob_filefind(const ob_type_filenum ddd):d(ddd){}
bool operator()(void* dd){
if(((ob_file*)dd)->getnum()==d) return true; else return false;
}
ob_type_hashtablelength gethash(){return (ob_type_hashtablelength)d;}
};

extern ob_type_namestatic ob_THENkwrd;
extern ob_type_namestatic ob_ENDIFkwrd;
extern ob_type_namestatic ob_TOkwrd;
extern ob_type_namestatic ob_STEPkwrd;
extern ob_type_namestatic ob_NEXTkwrd;
extern ob_type_namestatic ob_FORkwrd;
extern ob_type_namestatic ob_INPUTkwrd;
extern ob_type_namestatic ob_OUTPUTkwrd;
extern ob_type_namestatic ob_ASkwrd;
extern ob_type_namestatic ob_FILEkwrd;
extern ob_type_namestatic ob_ELSEkwrd;
extern ob_type_namestatic ob_DATAkwrd;
extern ob_type_namestatic ob_DUMMYkwrd;
extern ob_type_namestatic ob_SHARPkwrd;
extern ob_type_namestatic ob_REMkwrd;

class ob_lex{
public:

enum typelex{
PLUS,MINUS,MUL,DIV,POWER,ASSIGN,LP,RP,STRING,SEMICOLON,COMMA,ENDPROGRAMM,EOL,
LESS,GREAT,SHARP,BREAKPOINT,
DECINTNUMBER,HEXINTNUMBER,FLOATNUMBER,
FUNC,OPERATOR,
OLDVARIABLE,OLDARRAY,NEWNAME,
BAD
};

private:

ob_obasic* ptrbasic;

typelex            currentlex;
ob_type_string     currentlexasstring;
ob_type_namestatic currentlexasnamestatic;

ob_operator*       currentoperator;//for find operator

ob_api*            currentfunction;
ob_varone*         currentvar;
ob_vararray*       currentarray;

ob_type_streampos  previousptr;
ob_type_streampos  firstptr;

//istream for program need open in bynary mode
istream* in;

enum typego{GOTO,GOSUB};
void opergo(typego tg);
void findeol();
void dopinput(ob_lex* lex);
//for version 1.50
void findkeywordforif(ob_type_namestatic* kwrd1,ob_type_namestatic* kwrd2);

//expression calculation
void power(ob_savewrapper* left);
void term(ob_savewrapper* left);
void prim(ob_savewrapper* left);

//for version 1.50
void findother(ob_type_char other);//for ob_lex find CR-LF

public:

ob_api*            getcurrentfunction(){return currentfunction;}  //for ob_obasic::run
ob_operator*       getcurrentoperator(){return currentoperator;}  //for ob_obasic::run

ob_lex(istream* i,ob_obasic* p);

void setbasicptr(ob_obasic* p){ptrbasic=p;}//do not valid this in init list ob_obasic

~ob_lex(){}

void expr(ob_savewrapper* left);

typelex get_lex();//token analizer

void setin(istream* i){
in=i;
if(in!=OB_NULL)
firstptr=previousptr=in->tellg();
else
firstptr=previousptr=ob_type_streampos();
}

istream* getin(){return in;}//for ob_obasic::run - check run without load

typelex gettype() const {return currentlex;}

ob_type_namestatic& getstring(){return currentlexasnamestatic;}

ob_type_streampos getptr() const {return in->tellg();}//for ob_par::work ob_obasic::load operforfor

void putbacklex(){ob_seekg(in,previousptr);}

void restore(){//for DATA operator
ob_seekg(in,firstptr);
get_lex();
}

void getsizearray(ob_type_dim* ptrdimcurrent);//for ob_lex::extractvar and ob_lex::operdim

ob_varbase* extractvar();//for ob_lex::operXXXX
ob_varbase* extractvarwithoutcreate();//for ob_obasic::readvar

void workvariable(ob_savewrapper* left);//for ob_obasic::run and ob_lex::operfor

void posit(const ob_label* lab){lab->posit(in);}
void posit(ob_type_streampos pos){ob_seekg(in,pos);}//for operreturn

void partostream(ostream* name,ob_label* one);//for ob_api creat string of all parameters

template <class ob_type>
typename ob_type::typedatavar gettypevalue(){//do not use for ob_str

ob_savewrapper tmp;
expr(&tmp);


typename ob_type::typedatavar ret;

tmp.getptr()->cast(&ret);

return ret;
}

//operators
void operprint();     //1
void operinput();     //2
void operif();        //3
void operfor();       //4
void opernext();      //5
void opergoto();      //6
void opergosub();     //7
void operreturn();    //8
void operlet();       //9
void operdim();       //10
void operstop();      //11
void operend();       //12
void operrem();       //13
void operopen();      //14
void operclose();     //15
void operkill();      //16
void operread();      //17
void operdata();      //18
void operrestore();   //19
void operrandomize(); //20
void operelse();      //21
void operendif();     //22

};//stop class ob_lex

#define OB_NUMTYPES 3

#define OB_TYPEFLO 0
#define OB_TYPEINT 1
#define OB_TYPESTR 2

class ob_par{
public:
ob_type_parnum* mo[OB_NUMTYPES];
ob_ostrstream *name,*pc;
ob_type_flo *pf;
ob_type_int *pi;
void initpar();
ob_par(){initpar();}
void work(ob_lex* p);
void deletepar();
~ob_par(){deletepar();}
};

//for API defines
#define OB_DECLFUNAPI(name,typeret)                      \
typeret name(                                            \
ob_obasic* ptr,                                          \
const ob_type_char* parstring,                           \
const ob_type_parnum *descrf,                            \
const ob_type_parnum *descri,                            \
const ob_type_parnum *descrc,                            \
const ob_type_flo* parf,                                 \
const ob_type_int* pari,                                 \
const ob_type_char* parc)

#define OB_TYPEFUNAPI(name,typeret)                      \
typedef typeret (*name)(                                 \
ob_obasic* ptr,                                          \
const ob_type_char* parstring,                           \
const ob_type_parnum *descrf,                            \
const ob_type_parnum *descri,                            \
const ob_type_parnum *descrc,                            \
const ob_type_flo* parf,                                 \
const ob_type_int* pari,                                 \
const ob_type_char* parc);

OB_TYPEFUNAPI(ob_type_flofun,ob_type_flo)
OB_TYPEFUNAPI(ob_type_intfun,ob_type_int)
OB_TYPEFUNAPI(ob_type_charfun,ob_type_char*)

enum ob_type_del{OB_DELETE,OB_NODELETE};

#define OB_CLASSAPI(name,typeptr,dopdata,construct)             \
class name:public ob_api{                                       \
typeptr ptrfun;                                                 \
dopdata                                                         \
public:                                                         \
typedef typeptr typefun;                                        \
construct;                                                      \
void run(ob_obasic* ptr,ob_par* p,ob_savewrapper* v);           \
~name(){}                                                       \
};

OB_CLASSAPI(ob_apif,ob_type_flofun,,ob_apif(ob_type_char* a,ob_type_flofun))
OB_CLASSAPI(ob_apii,ob_type_intfun,,ob_apii(ob_type_char* a,ob_type_intfun))
OB_CLASSAPI(ob_apic,ob_type_charfun,ob_type_del dt;,
ob_apic(ob_type_char* a,ob_type_charfun,ob_type_del d))

extern const ob_type_char* ob_englmessage[];

typedef mlist<void*,ob_type_countlist>                         ob_typelist;
typedef mhash<void*,ob_type_hashtablelength,ob_type_countlist> ob_typetable;
typedef mstack<ob_type_streampos,ob_type_countlist>            ob_typestack;

template<class typestore,class find>
bool
replaceobjcommon1(ob_typetable* table,typestore* tmpptr1,find& f){

void* tmpptr=OB_NULL;
void* tmpptr2=tmpptr1;
bool fl=table->replaceobj(f,tmpptr2,&(tmpptr));
if(fl==true) delete ((typestore*)tmpptr);
return fl;
}

template<class typecompare,class typestore>
bool
replaceobjcommon2(ob_typetable* table,typestore* tmpptr1,typecompare* name){

ob_hashfind<typecompare,typestore> find(*name);

return
replaceobjcommon1<typestore,ob_hashfind<typecompare,typestore> >(
table,tmpptr1,find);
}

template<class T>
void ob_clearlist(ob_typelist* list){
ob_type_countlist       j;

ob_type_countlist       lenlist=list->getcount();

void* delptr;
for(j=0;j!=lenlist;++j){
list->get(&delptr);
T* delptr1=(T*)delptr;
delete delptr1;
}//for

}

template<class T>
void ob_cleartable(ob_typetable* table){

ob_type_hashtablelength i;

ob_type_hashtablelength lenvect=table->getcount();

for(i=0;i!=lenvect;++i){
ob_clearlist<T>(&((*table)[i]));
}//for

}

template <class T>
T* ob_findcommon(ob_type_namestatic* a,ob_typetable* table1){

ob_hashfind<ob_type_namestatic,T> findcomm(*a);
void** voidptr=table1->findobj(findcomm);
T* currentptr;
if(voidptr!=OB_NULL) currentptr=(T*)(*voidptr); else currentptr=OB_NULL;
return currentptr;
}

template <class T> T* ob_findcommon(ob_type_char* name,ob_typetable* table1){
ob_type_namestatic a(name);
return ob_findcommon<T>(&a,table1);
}

class ob_functoraccessarray{
ob_type_arraysize* as;
public:
ob_functoraccessarray(ob_type_arraysize* as1):as(as1){}
ob_savewrapper* operator()(ob_obasic* ptr,ob_type_char* name);
};

template <bool flcreate>
class ob_functoraccessvar{
public:
ob_savewrapper* operator()(ob_obasic* ptr,ob_type_char* name){

istrstream tmp(name);

ob_lex curr(&tmp,ptr);

curr.get_lex();

ob_varbase* vb;

if(flcreate==false)
vb=curr.extractvarwithoutcreate();//may be return NULL pointer
else
vb=curr.extractvar();//may be return NULL pointer

if(vb==OB_NULL)
return OB_NULL;
else
return vb->get();
}

};

class ob_obasic{


private:

ob_obasic& operator=(ob_obasic& ob){return *this;}

ob_obasic(ob_obasic& ob):
tablevar(ob_hashlength),   tablearray(ob_hashlength),
tablefun(ob_hashlength),   tablelabel(ob_hashlength),
tableforfor(ob_hashlength),tablefile(ob_hashlength),
baselex(OB_NULL,OB_NULL),readlex(OB_NULL,OB_NULL),datalex(OB_NULL,OB_NULL),
datastream(OB_NULL)
{}

friend class ob_lex;
friend class ob_functoraccessarray;

public:

enum typeload{OVERLAY,NOOVERLAY};
enum step{NOSTEP,STEP};
enum typeend{
DUMMYEND,ENDFILEEND,ENDOPERATORDETECT,EOLEND,BREAKPOINTEND,NOLOADEND,
BREAKBEFOREOPERATOR,BREAKAFTEROPERATOR
};

step stepflag;

private:
//static variable
static const ob_type_char* aboutmsg;
//for version 1.50
static const ob_type_serialnum sernum;
static const ob_type_version ver;

//non-static variable
typeend te;//for return value. need member for reenter after bef/aft oper break
ob_type_streampos retvalueptr;
bool endoperatordetect;//for exit if END operator detect
void setendoperatordetect(bool fl){endoperatordetect=fl;}

//containers
ob_operator  tableoperator[ob_maxnumoperator];

ob_typetable tablevar;   //ob_varone
ob_typetable tablearray; //ob_vararray
ob_typetable tablefun;   //ob_savewrapperapi
ob_typetable tablelabel; //ob_label
ob_typetable tableforfor;//ob_forfor
ob_typetable tablefile;  //ob_file

//ob_typelist  listfile;
ob_typestack stackgosub;

//variables
ob_lex baselex;
ob_lex readlex;
ob_lex datalex;
ostream* out;//out for PRINT operator (need open in text mode)
ob_strstream* datastream;//for DATA operator, create in load

time_t starttime;
time_t stoptime;

//for version 1.50
ob_file* findfile(ob_type_filenum number);//for ob_lex::operXXXX

template <class ob_type,class ob_functoraccess>
bool readvarcommon(
ob_type_char *name,ob_savewrapper* aa,ob_functoraccess& functoraccess){
bool ret;
ob_savewrapper* aaa;
if((aaa=functoraccess(this,name))!=OB_NULL){
ob_savewrapper aaaa(aaa->getptr()->duplicate());
aa->setptr(new ob_type());
aa->asg(&aaaa);
ret=true;
}//if
else{
ret=false;
}//else
return ret;
}

template <class ob_type>
bool readvar1(
ob_type_char* name,
typename ob_type::typedatavar* val,
ob_type_arraysize* as=OB_NULL){

ob_functoraccessvar<false> functoraccess1;
ob_functoraccessarray functoraccess2(as);
ob_savewrapper aa;

bool fl;
if(as==OB_NULL)
fl=readvarcommon<ob_type>(name,&aa,functoraccess1);
else
fl=readvarcommon<ob_type>(name,&aa,functoraccess2);

if(fl==true) aa.getptr()->cast(val);
return fl;
}

template <class ob_type,class ob_functoraccess>
bool writevarcommon(
ob_type_char *name,
typename ob_type::typedatavar val,
ob_functoraccess& functoraccess){
bool ret;
ob_type* a=new ob_type(val);
ob_savewrapper aa(a);
ob_savewrapper* aaa;
if((aaa=functoraccess(this,name))!=OB_NULL){
aaa->asg(&aa);
ret=true;
}//if
else{
ret=false;
}//else
return ret;
}

template <class ob_type>
bool writevar1(
ob_type_char* name,
typename ob_type::typedatavar val,
ob_type_arraysize* as){

ob_functoraccessvar<true> functoraccess1;
ob_functoraccessarray functoraccess2(as);
if(as==OB_NULL)
return writevarcommon<ob_type>(name,val,functoraccess1);
else
return writevarcommon<ob_type>(name,val,functoraccess2);
}

template <class ob_functoraccess>
ob_type_ident strlenvarcommon(
ob_type_char* name,
ob_type_stringsize* len,
ob_functoraccess& functoraccess){

ob_type_ident ret=OB_NOIDENT;

ob_savewrapper* aaa;

if((aaa=functoraccess(this,name))!=OB_NULL){
ret=aaa->getptr()->isA();
*len=aaa->getptr()->size();
}//if

return ret;
}

template <class Tobapi>
bool setfun1(ob_type_char* name,typename Tobapi::typefun f){
ob_type_namestatic a(name);

return
replaceobjcommon2<ob_type_namestatic,ob_api>(&tablefun,new Tobapi(name,f),&a);
}

ob_operator* findoperator(ob_type_namestatic* namestatic);//setbreak and get_lex

ob_operator* tmpoperator;//for run operator because baselex::currentoperator can overwrite

void
createarray(ob_type_namestatic* namearraystatic,ob_type_dim* ptrdimcurrent){

replaceobjcommon2<ob_type_namestatic,ob_vararray>(
&tablearray,
new ob_vararray(*namearraystatic,ptrdimcurrent),
namearraystatic
);
}

public:

ob_obasic(istream* i=OB_NULL,ostream* o=&cout,istream* read=&cin);

~ob_obasic(){
clrtablfun();
clrtabl(NOOVERLAY);
}

time_t getstarttime(){return starttime;}
time_t getstoptime(){return stoptime;}

static const ob_type_char* about(){return aboutmsg;}

void load(istream* i,typeload a=ob_obasic::NOOVERLAY);

typeend run();

void setstep(step ts){stepflag=ts;}
step getstep() const {return stepflag;}

//for API variable and array
ob_type_ident typevar(ob_type_char* name){
ob_varone* current=ob_findcommon<ob_varone>(name,&tablevar);
if(current!=OB_NULL) return current->isA(); else return OB_NOIDENT;
}

ob_type_ident typearray(ob_type_char* name){
ob_vararray* current=ob_findcommon<ob_vararray>(name,&tablearray);
if(current!=OB_NULL) return current->isA(); else return OB_NOIDENT;

}

void createvar(ob_type_char* name){
ob_functoraccessvar<true> functoraccess;
functoraccess(this,name);
}

void createarray(
ob_type_char* name,ob_type_arraydimension kr,ob_type_arraysize* as){

ob_type_namestatic namearraystatic(name);
ob_type_dim ptrdimcurrent(as,kr);
createarray(&namearraystatic,&ptrdimcurrent);
}

ob_type_ident
strlenvar(
ob_type_char* name,
ob_type_stringsize* len,
ob_type_arraysize* as=OB_NULL){

ob_functoraccessvar<false> functoraccess1;
ob_functoraccessarray functoraccess2(as);

if(as==OB_NULL)
return strlenvarcommon(name,len,functoraccess1);
else
return strlenvarcommon(name,len,functoraccess2);
}

bool
writevar(ob_type_char* name,ob_type_flo val,ob_type_arraysize* as=OB_NULL){
return writevar1<ob_float>(name,val,as);
}

bool
writevar(ob_type_char* name,ob_type_int val,ob_type_arraysize* as=OB_NULL){
return writevar1<ob_int>(name,val,as);
}

bool
writevar(ob_type_char* name,ob_type_char* val,ob_type_arraysize* as=OB_NULL){
return writevar1<ob_str>(name,val,as);
}

bool
readvar(ob_type_char* name,ob_type_flo* val,ob_type_arraysize* as=OB_NULL){
return readvar1<ob_float>(name,val,as);
}

bool
readvar(ob_type_char* name,ob_type_int* val,ob_type_arraysize* as=OB_NULL){
return readvar1<ob_int>(name,val,as);
}

bool
readvar(ob_type_char* name,ob_type_char* val,ob_type_arraysize* as=OB_NULL){

ob_functoraccessvar<false> functoraccess1;
ob_functoraccessarray functoraccess2(as);

ob_savewrapper aa;
ob_type_char* val1;

bool fl;
if(as==OB_NULL)
fl=readvarcommon<ob_str>(name,&aa,functoraccess1);
else
fl=readvarcommon<ob_str>(name,&aa,functoraccess2);

if(fl==true){
aa.getptr()->cast(&val1);
ob_strcpy(val,val1);
}//if
return fl;
}

const ob_type_arraysize* getarraysize(
ob_type_char* name,ob_type_arraydimension* kr){

ob_vararray* current=ob_findcommon<ob_vararray>(name,&tablearray);

if(current!=OB_NULL)
return current->getsize(kr);
else
return OB_NULL;
}

//for API user function
bool setfun(ob_type_char* name,ob_type_flofun f){
return setfun1<ob_apif>(name,f);
}

bool setfun(ob_type_char* name,ob_type_intfun f){
return setfun1<ob_apii>(name,f);
}

bool setfun(ob_type_char* name,ob_type_charfun f,ob_type_del dt){
ob_type_namestatic a(name);

return
replaceobjcommon2<ob_type_namestatic,ob_api>
(&tablefun,new ob_apic(name,f,dt),&a);
}

//for version 1.50
bool delfun(ob_type_char* name){

ob_type_namestatic a(name);

ob_hashfind<ob_type_namestatic,ob_api> find(a);
void* delptr=OB_NULL;
bool ret=tablefun.get(find,&(delptr));
if(ret==true) delete ((ob_api*)delptr);
return ret;
}

static ob_type_serialnum serial_number(){return sernum;}
static ob_type_int version(){return ver;}

void clrtabl(typeload a=NOOVERLAY){

ob_cleartable<ob_label>(&tablelabel);
ob_cleartable<ob_forfor>(&tableforfor);
stackgosub.clear();

if(a==NOOVERLAY){
   if(datastream!=NULL) delete datastream;
   datastream=NULL;
   ob_cleartable<ob_varone>(&tablevar);
   ob_cleartable<ob_vararray>(&tablearray);
   ob_cleartable<ob_file>(&tablefile);
}//if

}

void clrtablfun(){ob_cleartable<ob_api>(&tablefun);}

bool setoperatorbreak(char* name,ob_operator::flbreak b,ob_operator::flbreak a);

ob_type_countlist getcounttablevar()  {
return tablevar.getcounttable<ob_type_countlist>();
}

ob_type_countlist getcounttablearray(){
return tablearray.getcounttable<ob_type_countlist>();
}

ob_type_countlist getcounttablefun()  {
return tablefun.getcounttable<ob_type_countlist>();
}

ob_type_countlist getcounttablelabel(){
return tablelabel.getcounttable<ob_type_countlist>();
}

ob_type_countlist getcounttablefor()  {
return tableforfor.getcounttable<ob_type_countlist>();
}

ob_type_countlist getcountstackgosub(){
return stackgosub.getcount();
}

ob_type_countlist getcountlistfile()  {
return tablefile.getcounttable<ob_type_countlist>();
}

ob_type_char* getcurrentoperatorname(){return tmpoperator->getdata();}

void setout(ostream* o){out=o;}
void setin(istream* i){readlex.setin(i);}

ob_lex::typelex gettypenextlex(){
baselex.get_lex();
return baselex.gettype();
}

void restoreinputpointer(){baselex.posit(retvalueptr);}

};//end class ob_obasic

class ob_err{
ob_type_codeerror code;                        
public:
ob_err(ob_type_codeerror c){code=c;}

virtual void release(ostream* os,const ob_type_char** currentmessage){
(*os)<<endl<<((currentmessage))[code-ob_basecodeerror]<<ends;
(*os)<<endl<<"code error= "<<(code-ob_basecodeerror)<<ends;
}

virtual ob_type_codeerror getcode(){return code;}
virtual ~ob_err(){}
};

//prototypes
void ob_allocspaceforvar(
ob_type_namestatic* str,ob_savewrapper* value,ob_type_arraysize size);//for varone and vararray

void ob_loadbreakstr(
istream* in,
ob_type_stringsize lengthbuffers,
ob_type_char* ptrbefore,
ob_type_char* ptrcurrent,
ob_type_char* ptrafter
);

ob_type_int ob_hexconvert(ob_type_char* a);

OB_DECLFUNAPI(ob_fun_rnd,ob_type_flo);
OB_DECLFUNAPI(ob_fun_abs,ob_type_flo);
OB_DECLFUNAPI(ob_fun_sin,ob_type_flo);
OB_DECLFUNAPI(ob_fun_cos,ob_type_flo);
OB_DECLFUNAPI(ob_fun_atn,ob_type_flo);
OB_DECLFUNAPI(ob_fun_sqr,ob_type_flo);
OB_DECLFUNAPI(ob_fun_exp,ob_type_flo);
OB_DECLFUNAPI(ob_fun_log,ob_type_flo);
OB_DECLFUNAPI(ob_fun_log10,ob_type_flo);
OB_DECLFUNAPI(ob_fun_int,ob_type_int);
OB_DECLFUNAPI(ob_fun_sgn,ob_type_int);
OB_DECLFUNAPI(ob_fun_len,ob_type_int);
OB_DECLFUNAPI(ob_fun_dat,ob_type_char*);
OB_DECLFUNAPI(ob_fun_clk,ob_type_char*);
OB_DECLFUNAPI(ob_fun_digit2str,ob_type_char*);
OB_DECLFUNAPI(str2flo,ob_type_flo);
OB_DECLFUNAPI(str2int,ob_type_int);
OB_DECLFUNAPI(ob_fun_digit2hexstr,ob_type_char*);

#endif

